###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import unittest

from autobahn.wamp import message, role
from autobahn.wamp.exception import InvalidUriError, ProtocolError


class Foo(object):
    pass


class TestIds(unittest.TestCase):
    def test_valid_ids(self):
        for val in [0, 1, 23, 100000, 9007199254740992]:
            self.assertEqual(val, message.check_or_raise_id(val))

    def test_invalid_ids(self):
        for val in [
            -1,
            -9007199254740992,
            None,
            b"",
            b"abc",
            "",
            "abc",
            0.9,
            Foo(),
            False,
            True,
            [],
            {},
        ]:
            self.assertRaises(ProtocolError, message.check_or_raise_id, val)


class TestUris(unittest.TestCase):
    def test_valid_uris_loose_nonempty(self):
        for u in [
            "com.myapp.topic1",
            "com.myapp.product.123",
            "com.myapp.product.1.delete",
            "Com-star.MyApp.**+$for",
            "\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5",
            "hello\x24world",
            "hello\xc2\xa2world",
            "hello\xe2\x82\xacworld",
            "hello\xf0\xa4\xad\xa2world",
        ]:
            self.assertEqual(u, message.check_or_raise_uri(u))

    def test_invalid_uris_loose_nonempty(self):
        for u in [
            0,
            None,
            True,
            False,
            0.8,
            b"abc",
            Foo(),
            "",
            ".",
            "com.",
            "com..product",
            "com.my app.product",
            "com.my\tapp.product",
            "com.my\napp.product",
            "com.myapp.product#",
            "com.#.product",
        ]:
            self.assertRaises(InvalidUriError, message.check_or_raise_uri, u)

    def test_valid_uris_loose_empty(self):
        for u in [
            "com.myapp.topic1",
            "com.myapp..123",
            "com.myapp.product.1.",
            "com.",
            ".",
            "",
            "Com-star.MyApp.**+$for..foo",
            "\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5..foo",
            "hello\x24world..foo",
            "hello\xc2\xa2world..foo",
            "hello\xe2\x82\xacworld..foo",
            "hello\xf0\xa4\xad\xa2world..foo",
        ]:
            self.assertEqual(
                u, message.check_or_raise_uri(u, allow_empty_components=True)
            )

    def test_invalid_uris_loose_empty(self):
        for u in [
            0,
            None,
            True,
            False,
            0.8,
            b"abc",
            Foo(),
            "com.my app.product",
            "com.my\tapp.product",
            "com.my\napp.product",
            "com.myapp.product#",
            "com.#.product",
        ]:
            self.assertRaises(
                InvalidUriError,
                message.check_or_raise_uri,
                u,
                allow_empty_components=True,
            )

    def test_valid_uris_strict_nonempty(self):
        for u in [
            "com.myapp.topic1",
            "com.myapp.product.123",
            "com.myapp.product.1.delete",
        ]:
            self.assertEqual(u, message.check_or_raise_uri(u, strict=True))

    def test_invalid_uris_strict_nonempty(self):
        for u in [
            0,
            None,
            True,
            False,
            0.8,
            b"abc",
            Foo(),
            "",
            ".",
            "com.",
            "com..product",
            "com.my app.product",
            "com.my\tapp.product",
            "com.my\napp.product",
            "com.myapp.product#",
            "com.#.product",
            "Com-star.MyApp.**+$for",
            "\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5",
            "hello\x24world",
            "hello\xc2\xa2world",
            "hello\xe2\x82\xacworld",
            "hello\xf0\xa4\xad\xa2world",
        ]:
            self.assertRaises(
                InvalidUriError, message.check_or_raise_uri, u, strict=True
            )

    def test_valid_uris_strict_empty(self):
        for u in [
            "com.myapp.topic1",
            "com.myapp..123",
            "com.myapp.product.1.",
            "com.",
            ".",
            "",
        ]:
            self.assertEqual(
                u,
                message.check_or_raise_uri(u, strict=True, allow_empty_components=True),
            )

    def test_invalid_uris_strict_empty(self):
        for u in [
            0,
            None,
            True,
            False,
            0.8,
            b"abc",
            Foo(),
            "com.my app.product",
            "com.my\tapp.product",
            "com.my\napp.product",
            "com.myapp.product#",
            "com.#.product",
            "Com-star.MyApp.**+$for..foo",
            "\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5..foo",
            "hello\x24world..foo",
            "hello\xc2\xa2world..foo",
            "hello\xe2\x82\xacworld..foo",
            "hello\xf0\xa4\xad\xa2world..foo",
        ]:
            self.assertRaises(
                InvalidUriError,
                message.check_or_raise_uri,
                u,
                strict=True,
                allow_empty_components=True,
            )


class TestErrorMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Error(message.Call.MESSAGE_TYPE, 123456, "com.myapp.error1")
        msg = e.marshal()
        self.assertEqual(len(msg), 5)
        self.assertEqual(msg[0], message.Error.MESSAGE_TYPE)
        self.assertEqual(msg[1], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[2], 123456)
        self.assertEqual(msg[3], {})
        self.assertEqual(msg[4], "com.myapp.error1")

        e = message.Error(
            message.Call.MESSAGE_TYPE,
            123456,
            "com.myapp.error1",
            args=[1, 2, 3],
            kwargs={"foo": 23, "bar": "hello"},
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 7)
        self.assertEqual(msg[0], message.Error.MESSAGE_TYPE)
        self.assertEqual(msg[1], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[2], 123456)
        self.assertEqual(msg[3], {})
        self.assertEqual(msg[4], "com.myapp.error1")
        self.assertEqual(msg[5], [1, 2, 3])
        self.assertEqual(msg[6], {"foo": 23, "bar": "hello"})

    def test_parse_and_marshal(self):
        wmsg = [
            message.Error.MESSAGE_TYPE,
            message.Call.MESSAGE_TYPE,
            123456,
            {},
            "com.myapp.error1",
        ]
        msg = message.Error.parse(wmsg)
        self.assertIsInstance(msg, message.Error)
        self.assertEqual(msg.request_type, message.Call.MESSAGE_TYPE)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.error, "com.myapp.error1")
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Error.MESSAGE_TYPE,
            message.Call.MESSAGE_TYPE,
            123456,
            {},
            "com.myapp.error1",
            [1, 2, 3],
            {"foo": 23, "bar": "hello"},
        ]
        msg = message.Error.parse(wmsg)
        self.assertIsInstance(msg, message.Error)
        self.assertEqual(msg.request_type, message.Call.MESSAGE_TYPE)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.error, "com.myapp.error1")
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {"foo": 23, "bar": "hello"})
        self.assertEqual(msg.marshal(), wmsg)


class TestSubscribeMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Subscribe(123456, "com.myapp.topic1")
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Subscribe.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], "com.myapp.topic1")

        e = message.Subscribe(
            123456, "com.myapp.topic1", match=message.Subscribe.MATCH_PREFIX
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Subscribe.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {"match": "prefix"})
        self.assertEqual(msg[3], "com.myapp.topic1")

    def test_parse_and_marshal(self):
        wmsg = [message.Subscribe.MESSAGE_TYPE, 123456, {}, "com.myapp.topic1"]
        msg = message.Subscribe.parse(wmsg)
        self.assertIsInstance(msg, message.Subscribe)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, "com.myapp.topic1")
        self.assertEqual(msg.match, message.Subscribe.MATCH_EXACT)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Subscribe.MESSAGE_TYPE,
            123456,
            {"match": "prefix"},
            "com.myapp.topic1",
        ]
        msg = message.Subscribe.parse(wmsg)
        self.assertIsInstance(msg, message.Subscribe)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, "com.myapp.topic1")
        self.assertEqual(msg.match, message.Subscribe.MATCH_PREFIX)
        self.assertEqual(msg.marshal(), wmsg)

    def test_get_retained_default_false(self):
        wmsg = [
            message.Subscribe.MESSAGE_TYPE,
            123456,
            {"match": "prefix"},
            "com.myapp.topic1",
        ]
        msg = message.Subscribe.parse(wmsg)
        self.assertIsInstance(msg, message.Subscribe)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, "com.myapp.topic1")
        self.assertEqual(msg.get_retained, None)
        self.assertNotEqual(msg.get_retained, True)
        self.assertEqual(msg.match, message.Subscribe.MATCH_PREFIX)
        self.assertEqual(msg.marshal(), wmsg)

    def test_get_retained_explicit_false(self):
        wmsg = [
            message.Subscribe.MESSAGE_TYPE,
            123456,
            {"match": "prefix", "get_retained": False},
            "com.myapp.topic1",
        ]
        msg = message.Subscribe.parse(wmsg)
        self.assertIsInstance(msg, message.Subscribe)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, "com.myapp.topic1")
        self.assertEqual(msg.get_retained, False)
        self.assertNotEqual(msg.get_retained, True)
        self.assertEqual(msg.match, message.Subscribe.MATCH_PREFIX)
        self.assertEqual(msg.marshal(), wmsg)

    def test_get_retained_explicit_true(self):
        wmsg = [
            message.Subscribe.MESSAGE_TYPE,
            123456,
            {"match": "prefix", "get_retained": True},
            "com.myapp.topic1",
        ]
        msg = message.Subscribe.parse(wmsg)
        self.assertIsInstance(msg, message.Subscribe)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, "com.myapp.topic1")
        self.assertEqual(msg.get_retained, True)
        self.assertEqual(msg.match, message.Subscribe.MATCH_PREFIX)
        self.assertEqual(msg.marshal(), wmsg)


class TestSubscribedMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Subscribed(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Subscribed.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Subscribed.MESSAGE_TYPE, 123456, 789123]
        msg = message.Subscribed.parse(wmsg)
        self.assertIsInstance(msg, message.Subscribed)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.subscription, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestUnsubscribeMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Unsubscribe(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Unsubscribe.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Unsubscribe.MESSAGE_TYPE, 123456, 789123]
        msg = message.Unsubscribe.parse(wmsg)
        self.assertIsInstance(msg, message.Unsubscribe)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.subscription, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestUnsubscribedMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Unsubscribed(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 2)
        self.assertEqual(msg[0], message.Unsubscribed.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)

        e = message.Unsubscribed(0, subscription=123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Unsubscribed.MESSAGE_TYPE)
        self.assertEqual(msg[1], 0)
        self.assertEqual(msg[2], {"subscription": 123456})

        e = message.Unsubscribed(
            0, subscription=123456, reason="wamp.subscription.revoked"
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Unsubscribed.MESSAGE_TYPE)
        self.assertEqual(msg[1], 0)
        self.assertEqual(
            msg[2], {"subscription": 123456, "reason": "wamp.subscription.revoked"}
        )

    def test_parse_and_marshal(self):
        wmsg = [message.Unsubscribed.MESSAGE_TYPE, 123456]
        msg = message.Unsubscribed.parse(wmsg)
        self.assertIsInstance(msg, message.Unsubscribed)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.subscription, None)
        self.assertEqual(msg.reason, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Unsubscribed.MESSAGE_TYPE, 0, {"subscription": 123456}]
        msg = message.Unsubscribed.parse(wmsg)
        self.assertIsInstance(msg, message.Unsubscribed)
        self.assertEqual(msg.request, 0)
        self.assertEqual(msg.subscription, 123456)
        self.assertEqual(msg.reason, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Unsubscribed.MESSAGE_TYPE,
            0,
            {"subscription": 123456, "reason": "wamp.subscription.revoked"},
        ]
        msg = message.Unsubscribed.parse(wmsg)
        self.assertIsInstance(msg, message.Unsubscribed)
        self.assertEqual(msg.request, 0)
        self.assertEqual(msg.subscription, 123456)
        self.assertEqual(msg.reason, "wamp.subscription.revoked")
        self.assertEqual(msg.marshal(), wmsg)


class TestPublishMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Publish(123456, "com.myapp.topic1")
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Publish.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], "com.myapp.topic1")

        e = message.Publish(
            123456,
            "com.myapp.topic1",
            args=[1, 2, 3],
            kwargs={"foo": 23, "bar": "hello"},
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 6)
        self.assertEqual(msg[0], message.Publish.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], "com.myapp.topic1")
        self.assertEqual(msg[4], [1, 2, 3])
        self.assertEqual(msg[5], {"foo": 23, "bar": "hello"})

        e = message.Publish(
            123456,
            "com.myapp.topic1",
            exclude_me=False,
            exclude=[300],
            eligible=[100, 200, 300],
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Publish.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(
            msg[2], {"exclude_me": False, "exclude": [300], "eligible": [100, 200, 300]}
        )
        self.assertEqual(msg[3], "com.myapp.topic1")

    def test_parse_and_marshal(self):
        wmsg = [message.Publish.MESSAGE_TYPE, 123456, {}, "com.myapp.topic1"]
        msg = message.Publish.parse(wmsg)
        self.assertIsInstance(msg, message.Publish)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, "com.myapp.topic1")
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.exclude_me, None)
        self.assertEqual(msg.exclude, None)
        self.assertEqual(msg.eligible, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Publish.MESSAGE_TYPE,
            123456,
            {},
            "com.myapp.topic1",
            [1, 2, 3],
            {"foo": 23, "bar": "hello"},
        ]
        msg = message.Publish.parse(wmsg)
        self.assertIsInstance(msg, message.Publish)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, "com.myapp.topic1")
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {"foo": 23, "bar": "hello"})
        self.assertEqual(msg.exclude_me, None)
        self.assertEqual(msg.exclude, None)
        self.assertEqual(msg.eligible, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Publish.MESSAGE_TYPE,
            123456,
            {"exclude_me": False, "exclude": [300], "eligible": [100, 200, 300]},
            "com.myapp.topic1",
        ]
        msg = message.Publish.parse(wmsg)
        self.assertIsInstance(msg, message.Publish)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.topic, "com.myapp.topic1")
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.exclude_me, False)
        self.assertEqual(msg.exclude, [300])
        self.assertEqual(msg.eligible, [100, 200, 300])
        self.assertEqual(msg.marshal(), wmsg)

    def test_retain_default_false(self):
        """
        Retain, when not specified, is False-y by default.
        """
        wmsg = [
            message.Publish.MESSAGE_TYPE,
            123456,
            {"exclude_me": False, "exclude": [300], "eligible": [100, 200, 300]},
            "com.myapp.topic1",
        ]
        msg = message.Publish.parse(wmsg)
        self.assertIsInstance(msg, message.Publish)
        self.assertEqual(msg.retain, None)
        self.assertIsNot(msg.retain, True)
        self.assertEqual(msg.marshal(), wmsg)

    def test_retain_explicit_false(self):
        """
        Retain, when specified as False, shows up in the message.
        """
        wmsg = [
            message.Publish.MESSAGE_TYPE,
            123456,
            {
                "exclude_me": False,
                "retain": False,
                "exclude": [300],
                "eligible": [100, 200, 300],
            },
            "com.myapp.topic1",
        ]
        msg = message.Publish.parse(wmsg)
        self.assertIsInstance(msg, message.Publish)
        self.assertEqual(msg.retain, False)
        self.assertIsNot(msg.retain, True)
        self.assertEqual(msg.marshal(), wmsg)

    def test_retain_explicit_true(self):
        """
        Retain, when specified as True, shows up in the message.
        """
        wmsg = [
            message.Publish.MESSAGE_TYPE,
            123456,
            {
                "exclude_me": False,
                "retain": True,
                "exclude": [300],
                "eligible": [100, 200, 300],
            },
            "com.myapp.topic1",
        ]
        msg = message.Publish.parse(wmsg)
        self.assertIsInstance(msg, message.Publish)
        self.assertEqual(msg.retain, True)
        self.assertIs(msg.retain, True)
        self.assertEqual(msg.marshal(), wmsg)


class TestPublishedMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Published(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Published.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Published.MESSAGE_TYPE, 123456, 789123]
        msg = message.Published.parse(wmsg)
        self.assertIsInstance(msg, message.Published)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.publication, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestEventMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Event(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Event.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {})

        e = message.Event(
            123456, 789123, args=[1, 2, 3], kwargs={"foo": 23, "bar": "hello"}
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 6)
        self.assertEqual(msg[0], message.Event.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {})
        self.assertEqual(msg[4], [1, 2, 3])
        self.assertEqual(msg[5], {"foo": 23, "bar": "hello"})

        e = message.Event(123456, 789123, publisher=300)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Event.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {"publisher": 300})

    def test_parse_and_marshal(self):
        wmsg = [message.Event.MESSAGE_TYPE, 123456, 789123, {}]
        msg = message.Event.parse(wmsg)
        self.assertIsInstance(msg, message.Event)
        self.assertEqual(msg.subscription, 123456)
        self.assertEqual(msg.publication, 789123)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.publisher, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Event.MESSAGE_TYPE,
            123456,
            789123,
            {},
            [1, 2, 3],
            {"foo": 23, "bar": "hello"},
        ]
        msg = message.Event.parse(wmsg)
        self.assertIsInstance(msg, message.Event)
        self.assertEqual(msg.subscription, 123456)
        self.assertEqual(msg.publication, 789123)
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {"foo": 23, "bar": "hello"})
        self.assertEqual(msg.publisher, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Event.MESSAGE_TYPE, 123456, 789123, {"publisher": 300}]
        msg = message.Event.parse(wmsg)
        self.assertIsInstance(msg, message.Event)
        self.assertEqual(msg.subscription, 123456)
        self.assertEqual(msg.publication, 789123)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.publisher, 300)
        self.assertEqual(msg.marshal(), wmsg)

    def test_retained_default_false(self):
        wmsg = [message.Event.MESSAGE_TYPE, 123456, 789123, {}]
        msg = message.Event.parse(wmsg)
        self.assertIsInstance(msg, message.Event)
        self.assertEqual(msg.retained, None)
        self.assertNotEqual(msg.retained, True)
        self.assertEqual(msg.marshal(), wmsg)

    def test_retained_explicit_false(self):
        wmsg = [message.Event.MESSAGE_TYPE, 123456, 789123, {"retained": False}]
        msg = message.Event.parse(wmsg)
        self.assertIsInstance(msg, message.Event)
        self.assertEqual(msg.retained, False)
        self.assertNotEqual(msg.retained, True)
        self.assertEqual(msg.marshal(), wmsg)

    def test_retained_explicit_true(self):
        wmsg = [message.Event.MESSAGE_TYPE, 123456, 789123, {"retained": True}]
        msg = message.Event.parse(wmsg)
        self.assertIsInstance(msg, message.Event)
        self.assertEqual(msg.retained, True)
        self.assertEqual(msg.marshal(), wmsg)


class TestRegisterMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Register(123456, "com.myapp.procedure1")
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Register.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], "com.myapp.procedure1")

        e = message.Register(123456, "com.myapp.procedure1", match="wildcard")
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Register.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {"match": "wildcard"})
        self.assertEqual(msg[3], "com.myapp.procedure1")

    def test_ctor_reregister(self):
        e = message.Register(123456, "com.myapp.procedure1", force_reregister=True)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Register.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {"force_reregister": True})
        self.assertEqual(msg[3], "com.myapp.procedure1")

        e2 = message.Register.parse(msg)
        str(e2)

    def test_parse_reregister_illegal_force(self):
        msg = [
            message.Register.MESSAGE_TYPE,
            123456,
            {"force_reregister": "truthy"},
            "com.myapp.procedure1",
        ]

        with self.assertRaises(ProtocolError) as ctx:
            message.Register.parse(msg)
        self.assertIn("invalid type", str(ctx.exception))

    def test_parse_and_marshal(self):
        wmsg = [message.Register.MESSAGE_TYPE, 123456, {}, "com.myapp.procedure1"]
        msg = message.Register.parse(wmsg)
        self.assertIsInstance(msg, message.Register)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, "com.myapp.procedure1")
        self.assertEqual(msg.match, "exact")
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Register.MESSAGE_TYPE,
            123456,
            {"match": "wildcard"},
            "com.myapp.procedure1",
        ]
        msg = message.Register.parse(wmsg)
        self.assertIsInstance(msg, message.Register)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, "com.myapp.procedure1")
        self.assertEqual(msg.match, "wildcard")
        self.assertEqual(msg.marshal(), wmsg)


class TestRegisteredMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Registered(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Registered.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Registered.MESSAGE_TYPE, 123456, 789123]
        msg = message.Registered.parse(wmsg)
        self.assertIsInstance(msg, message.Registered)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestUnregisterMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Unregister(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Unregister.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)

    def test_parse_and_marshal(self):
        wmsg = [message.Unregister.MESSAGE_TYPE, 123456, 789123]
        msg = message.Unregister.parse(wmsg)
        self.assertIsInstance(msg, message.Unregister)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.marshal(), wmsg)


class TestUnregisteredMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Unregistered(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 2)
        self.assertEqual(msg[0], message.Unregistered.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)

        e = message.Unregistered(0, registration=123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Unregistered.MESSAGE_TYPE)
        self.assertEqual(msg[1], 0)
        self.assertEqual(msg[2], {"registration": 123456})

        e = message.Unregistered(
            0, registration=123456, reason="wamp.registration.revoked"
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Unregistered.MESSAGE_TYPE)
        self.assertEqual(msg[1], 0)
        self.assertEqual(
            msg[2], {"registration": 123456, "reason": "wamp.registration.revoked"}
        )

    def test_parse_and_marshal(self):
        wmsg = [message.Unregistered.MESSAGE_TYPE, 123456]
        msg = message.Unregistered.parse(wmsg)
        self.assertIsInstance(msg, message.Unregistered)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, None)
        self.assertEqual(msg.reason, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Unregistered.MESSAGE_TYPE, 0, {"registration": 123456}]
        msg = message.Unregistered.parse(wmsg)
        self.assertIsInstance(msg, message.Unregistered)
        self.assertEqual(msg.request, 0)
        self.assertEqual(msg.registration, 123456)
        self.assertEqual(msg.reason, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Unregistered.MESSAGE_TYPE,
            0,
            {"registration": 123456, "reason": "wamp.registration.revoked"},
        ]
        msg = message.Unregistered.parse(wmsg)
        self.assertIsInstance(msg, message.Unregistered)
        self.assertEqual(msg.request, 0)
        self.assertEqual(msg.registration, 123456)
        self.assertEqual(msg.reason, "wamp.registration.revoked")
        self.assertEqual(msg.marshal(), wmsg)


class TestCallMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Call(123456, "com.myapp.procedure1")
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], "com.myapp.procedure1")

        e = message.Call(
            123456,
            "com.myapp.procedure1",
            args=[1, 2, 3],
            kwargs={"foo": 23, "bar": "hello"},
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 6)
        self.assertEqual(msg[0], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], "com.myapp.procedure1")
        self.assertEqual(msg[4], [1, 2, 3])
        self.assertEqual(msg[5], {"foo": 23, "bar": "hello"})

        e = message.Call(123456, "com.myapp.procedure1", timeout=10000)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Call.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {"timeout": 10000})
        self.assertEqual(msg[3], "com.myapp.procedure1")

    def test_parse_and_marshal(self):
        wmsg = [message.Call.MESSAGE_TYPE, 123456, {}, "com.myapp.procedure1"]
        msg = message.Call.parse(wmsg)
        self.assertIsInstance(msg, message.Call)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, "com.myapp.procedure1")
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.timeout, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Call.MESSAGE_TYPE,
            123456,
            {},
            "com.myapp.procedure1",
            [1, 2, 3],
            {"foo": 23, "bar": "hello"},
        ]
        msg = message.Call.parse(wmsg)
        self.assertIsInstance(msg, message.Call)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, "com.myapp.procedure1")
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {"foo": 23, "bar": "hello"})
        self.assertEqual(msg.timeout, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Call.MESSAGE_TYPE,
            123456,
            {"timeout": 10000},
            "com.myapp.procedure1",
        ]
        msg = message.Call.parse(wmsg)
        self.assertIsInstance(msg, message.Call)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.procedure, "com.myapp.procedure1")
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.timeout, 10000)
        self.assertEqual(msg.marshal(), wmsg)


class TestCancelMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Cancel(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Cancel.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})

        e = message.Cancel(123456, mode=message.Cancel.KILL)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Cancel.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {"mode": message.Cancel.KILL})

    def test_parse_and_marshal(self):
        wmsg = [message.Cancel.MESSAGE_TYPE, 123456, {}]
        msg = message.Cancel.parse(wmsg)
        self.assertIsInstance(msg, message.Cancel)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.mode, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Cancel.MESSAGE_TYPE, 123456, {"mode": message.Cancel.KILL}]
        msg = message.Cancel.parse(wmsg)
        self.assertIsInstance(msg, message.Cancel)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.mode, message.Cancel.KILL)
        self.assertEqual(msg.marshal(), wmsg)


class TestResultMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Result(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Result.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})

        e = message.Result(123456, args=[1, 2, 3], kwargs={"foo": 23, "bar": "hello"})
        msg = e.marshal()
        self.assertEqual(len(msg), 5)
        self.assertEqual(msg[0], message.Result.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], [1, 2, 3])
        self.assertEqual(msg[4], {"foo": 23, "bar": "hello"})

        e = message.Result(123456, progress=True)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Result.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {"progress": True})

    def test_parse_and_marshal(self):
        wmsg = [message.Result.MESSAGE_TYPE, 123456, {}]
        msg = message.Result.parse(wmsg)
        self.assertIsInstance(msg, message.Result)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.progress, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Result.MESSAGE_TYPE,
            123456,
            {},
            [1, 2, 3],
            {"foo": 23, "bar": "hello"},
        ]
        msg = message.Result.parse(wmsg)
        self.assertIsInstance(msg, message.Result)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {"foo": 23, "bar": "hello"})
        self.assertEqual(msg.progress, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Result.MESSAGE_TYPE, 123456, {"progress": True}]
        msg = message.Result.parse(wmsg)
        self.assertIsInstance(msg, message.Result)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.progress, True)
        self.assertEqual(msg.marshal(), wmsg)


class TestInvocationMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Invocation(123456, 789123)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Invocation.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {})

        e = message.Invocation(
            123456, 789123, args=[1, 2, 3], kwargs={"foo": 23, "bar": "hello"}
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 6)
        self.assertEqual(msg[0], message.Invocation.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {})
        self.assertEqual(msg[4], [1, 2, 3])
        self.assertEqual(msg[5], {"foo": 23, "bar": "hello"})

        e = message.Invocation(123456, 789123, timeout=10000)
        msg = e.marshal()
        self.assertEqual(len(msg), 4)
        self.assertEqual(msg[0], message.Invocation.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], 789123)
        self.assertEqual(msg[3], {"timeout": 10000})

    def test_parse_and_marshal(self):
        wmsg = [message.Invocation.MESSAGE_TYPE, 123456, 789123, {}]
        msg = message.Invocation.parse(wmsg)
        self.assertIsInstance(msg, message.Invocation)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.timeout, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Invocation.MESSAGE_TYPE,
            123456,
            789123,
            {},
            [1, 2, 3],
            {"foo": 23, "bar": "hello"},
        ]
        msg = message.Invocation.parse(wmsg)
        self.assertIsInstance(msg, message.Invocation)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {"foo": 23, "bar": "hello"})
        self.assertEqual(msg.timeout, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Invocation.MESSAGE_TYPE, 123456, 789123, {"timeout": 10000}]
        msg = message.Invocation.parse(wmsg)
        self.assertIsInstance(msg, message.Invocation)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.registration, 789123)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.timeout, 10000)
        self.assertEqual(msg.marshal(), wmsg)


class TestInterruptMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Interrupt(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Interrupt.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})

        e = message.Interrupt(123456, mode=message.Interrupt.KILL)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Interrupt.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {"mode": message.Interrupt.KILL})

    def test_parse_and_marshal(self):
        wmsg = [message.Interrupt.MESSAGE_TYPE, 123456, {}]
        msg = message.Interrupt.parse(wmsg)
        self.assertIsInstance(msg, message.Interrupt)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.mode, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Interrupt.MESSAGE_TYPE,
            123456,
            {"mode": message.Interrupt.KILL},
        ]
        msg = message.Interrupt.parse(wmsg)
        self.assertIsInstance(msg, message.Interrupt)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.mode, message.Interrupt.KILL)
        self.assertEqual(msg.marshal(), wmsg)


class TestYieldMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Yield(123456)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Yield.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})

        e = message.Yield(123456, args=[1, 2, 3], kwargs={"foo": 23, "bar": "hello"})
        msg = e.marshal()
        self.assertEqual(len(msg), 5)
        self.assertEqual(msg[0], message.Yield.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {})
        self.assertEqual(msg[3], [1, 2, 3])
        self.assertEqual(msg[4], {"foo": 23, "bar": "hello"})

        e = message.Yield(123456, progress=True)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Yield.MESSAGE_TYPE)
        self.assertEqual(msg[1], 123456)
        self.assertEqual(msg[2], {"progress": True})

    def test_parse_and_marshal(self):
        wmsg = [message.Yield.MESSAGE_TYPE, 123456, {}]
        msg = message.Yield.parse(wmsg)
        self.assertIsInstance(msg, message.Yield)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.progress, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Yield.MESSAGE_TYPE,
            123456,
            {},
            [1, 2, 3],
            {"foo": 23, "bar": "hello"},
        ]
        msg = message.Yield.parse(wmsg)
        self.assertIsInstance(msg, message.Yield)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, [1, 2, 3])
        self.assertEqual(msg.kwargs, {"foo": 23, "bar": "hello"})
        self.assertEqual(msg.progress, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Yield.MESSAGE_TYPE, 123456, {"progress": True}]
        msg = message.Yield.parse(wmsg)
        self.assertIsInstance(msg, message.Yield)
        self.assertEqual(msg.request, 123456)
        self.assertEqual(msg.args, None)
        self.assertEqual(msg.kwargs, None)
        self.assertEqual(msg.progress, True)
        self.assertEqual(msg.marshal(), wmsg)


class TestHelloMessage(unittest.TestCase):
    def test_ctor(self):
        e = message.Hello("realm1", {"publisher": role.RolePublisherFeatures()})
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Hello.MESSAGE_TYPE)
        self.assertEqual(msg[1], "realm1")
        self.assertEqual(msg[2], {"roles": {"publisher": {}}})

        e = message.Hello(
            "realm1",
            {
                "publisher": role.RolePublisherFeatures(
                    subscriber_blackwhite_listing=True
                )
            },
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Hello.MESSAGE_TYPE)
        self.assertEqual(msg[1], "realm1")
        self.assertEqual(
            msg[2],
            {
                "roles": {
                    "publisher": {"features": {"subscriber_blackwhite_listing": True}}
                }
            },
        )

        e = message.Hello(
            "realm1",
            {
                "publisher": role.RolePublisherFeatures(
                    subscriber_blackwhite_listing=True
                )
            },
            resumable=True,
        )
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Hello.MESSAGE_TYPE)
        self.assertEqual(msg[1], "realm1")
        self.assertEqual(
            msg[2],
            {
                "roles": {
                    "publisher": {"features": {"subscriber_blackwhite_listing": True}}
                },
                "resumable": True,
            },
        )

    def test_parse_and_marshal(self):
        wmsg = [message.Hello.MESSAGE_TYPE, "realm1", {"roles": {"publisher": {}}}]
        msg = message.Hello.parse(wmsg)
        self.assertIsInstance(msg, message.Hello)
        self.assertEqual(msg.realm, "realm1")
        self.assertEqual(msg.roles, {"publisher": role.RolePublisherFeatures()})
        self.assertEqual(msg.resumable, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Hello.MESSAGE_TYPE,
            "realm1",
            {
                "roles": {
                    "publisher": {"features": {"subscriber_blackwhite_listing": True}}
                }
            },
        ]
        msg = message.Hello.parse(wmsg)
        self.assertIsInstance(msg, message.Hello)
        self.assertEqual(msg.realm, "realm1")
        self.assertEqual(
            msg.roles,
            {
                "publisher": role.RolePublisherFeatures(
                    subscriber_blackwhite_listing=True
                )
            },
        )
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Hello.MESSAGE_TYPE,
            "realm1",
            {"roles": {"publisher": {}}, "resumable": False},
        ]
        msg = message.Hello.parse(wmsg)
        self.assertIsInstance(msg, message.Hello)
        self.assertEqual(msg.realm, "realm1")
        self.assertEqual(msg.roles, {"publisher": role.RolePublisherFeatures()})
        self.assertEqual(msg.resumable, False)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [
            message.Hello.MESSAGE_TYPE,
            "realm1",
            {
                "roles": {"publisher": {}},
                "resumable": True,
                "resume-session": 1234,
                "resume-token": "dsjgsg",
            },
        ]
        msg = message.Hello.parse(wmsg)
        self.assertIsInstance(msg, message.Hello)
        self.assertEqual(msg.realm, "realm1")
        self.assertEqual(msg.roles, {"publisher": role.RolePublisherFeatures()})
        self.assertEqual(msg.resumable, True)
        self.assertEqual(msg.resume_session, 1234)
        self.assertEqual(msg.resume_token, "dsjgsg")
        self.assertEqual(msg.marshal(), wmsg)

    def test_str(self):
        e = message.Hello("realm1", {"publisher": role.RolePublisherFeatures()})
        self.assertIsInstance(str(e), str)


class TestGoodbyeMessage(unittest.TestCase):
    def test_ctor(self):
        reason = "wamp.error.system_shutdown"
        reason_msg = "The host is shutting down now."

        e = message.Goodbye()
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Goodbye.MESSAGE_TYPE)
        self.assertEqual(msg[1], {})
        self.assertEqual(msg[2], message.Goodbye.DEFAULT_REASON)

        e = message.Goodbye(reason=reason)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Goodbye.MESSAGE_TYPE)
        self.assertEqual(msg[1], {})
        self.assertEqual(msg[2], reason)

        e = message.Goodbye(reason=reason, message=reason_msg)
        msg = e.marshal()
        self.assertEqual(len(msg), 3)
        self.assertEqual(msg[0], message.Goodbye.MESSAGE_TYPE)
        self.assertEqual(msg[1], {"message": reason_msg})
        self.assertEqual(msg[2], reason)

    def test_parse_and_marshal(self):
        reason = "wamp.error.system_shutdown"
        reason_msg = "The host is shutting down now."

        wmsg = [message.Goodbye.MESSAGE_TYPE]
        self.assertRaises(ProtocolError, message.Goodbye.parse, wmsg)

        wmsg = [message.Goodbye.MESSAGE_TYPE, reason]
        self.assertRaises(ProtocolError, message.Goodbye.parse, wmsg)

        wmsg = [message.Goodbye.MESSAGE_TYPE, {"message": 100}, reason]
        self.assertRaises(ProtocolError, message.Goodbye.parse, wmsg)

        wmsg = [message.Goodbye.MESSAGE_TYPE, {}, reason]
        msg = message.Goodbye.parse(wmsg)
        self.assertIsInstance(msg, message.Goodbye)
        self.assertEqual(msg.reason, reason)
        self.assertEqual(msg.message, None)
        self.assertEqual(msg.marshal(), wmsg)

        wmsg = [message.Goodbye.MESSAGE_TYPE, {"message": reason_msg}, reason]
        msg = message.Goodbye.parse(wmsg)
        self.assertIsInstance(msg, message.Goodbye)
        self.assertEqual(msg.reason, reason)
        self.assertEqual(msg.message, reason_msg)
        self.assertEqual(msg.marshal(), wmsg)

    def test_str(self):
        e = message.Goodbye(
            reason="wamp.error.system_shutdown",
            message="The host is shutting down now.",
        )
        self.assertIsInstance(str(e), str)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

"""
Centralized compiler flag selection for NVX (Native Vector Extensions) builds.

This module determines appropriate architecture-specific optimization flags
for building native C extensions using CFFI. It balances portability and
performance based on the build context.

Strategy
--------

For **WHEEL BUILDS** (distribution via PyPI):
    Use safe, portable baseline architectures to ensure wheels work on a wide
    range of CPUs without causing SIGILL (Illegal Instruction) crashes.

For **SOURCE BUILDS** (local installation):
    Use -march=native to generate optimal code for the specific CPU where
    the build is happening, maximizing performance.

Architecture Baselines
----------------------

x86-64 (portable baseline):
    -march=x86-64-v2 (microarchitecture level 2, 2009+)
    Includes: SSE4.2, POPCNT, SSSE3, SSE4.1
    Compatible with: Intel Nehalem+, AMD Bulldozer+ (2009+)
    Coverage: ~99% of x86-64 CPUs from 2010 onwards

ARM64 (portable baseline):
    -march=armv8-a (baseline ARMv8-A architecture)
    Compatible with: Raspberry Pi 4/5, AWS Graviton, Apple M1/M2, etc.
    Coverage: All 64-bit ARM CPUs

Environment Variables
---------------------

AUTOBAHN_ARCH_TARGET : str, optional
    User/distro override for architecture target:
    - "native" : Force -march=native (maximum performance, may break portability)
    - "safe"   : Force portable baseline (ensures compatibility)
    - Not set  : Auto-detect based on build context (recommended)

AUTOBAHN_WHEEL_BUILD : str, optional
    Explicit marker for wheel builds ("true" or "1")
    When set, forces portable baseline regardless of auto-detection

CI : str, optional
    Standard CI environment variable. When "true" or "1", assumes wheel build.

CIBUILDWHEEL : str, optional
    Set by cibuildwheel. Indicates wheel build for distribution.

AUDITWHEEL_PLAT : str, optional
    Set by auditwheel/manylinux builds. Indicates wheel build.

Examples
--------

**GitHub Actions building wheels:**
    >>> # Automatically detects CI=true, uses -march=x86-64-v2
    >>> get_compile_args()
    ['-std=c99', '-Wall', '-Wno-strict-prototypes', '-O3', '-march=x86-64-v2']

**User installing from source:**
    >>> # Detects local build, uses -march=native
    >>> get_compile_args()
    ['-std=c99', '-Wall', '-Wno-strict-prototypes', '-O3', '-march=native']

**Gentoo building package:**
    >>> # Gentoo wants -march=native for optimized binaries
    >>> import os
    >>> os.environ['AUTOBAHN_ARCH_TARGET'] = 'native'
    >>> get_compile_args()
    ['-std=c99', '-Wall', '-Wno-strict-prototypes', '-O3', '-march=native']

**Debian building package:**
    >>> # Debian wants portable binaries for all users
    >>> import os
    >>> os.environ['AUTOBAHN_ARCH_TARGET'] = 'safe'
    >>> get_compile_args()
    ['-std=c99', '-Wall', '-Wno-strict-prototypes', '-O3', '-march=x86-64-v2']

Notes
-----

This module is used by all NVX components:
- autobahn.nvx._xormasker (WebSocket frame XOR masking)
- autobahn.nvx._utf8validator (WebSocket UTF-8 validation)

Related Issues
--------------
- #1717: SIGILL crashes from -march=native in distributed wheels

See Also
--------
- https://en.wikipedia.org/wiki/X86-64#Microarchitecture_levels
- https://gcc.gnu.org/onlinedocs/gcc/x86-Options.html
"""

import os
import sys
import platform


def is_building_wheel():
    """
    Detect if we're building a wheel for distribution vs. a local source install.

    Returns
    -------
    bool
        True if building a wheel (need portability), False if building locally.

    Notes
    -----
    Detection heuristics:
    1. CI environment variables (CI, CIBUILDWHEEL)
    2. manylinux/auditwheel markers (AUDITWHEEL_PLAT)
    3. Explicit marker (AUTOBAHN_WHEEL_BUILD)
    4. Default: False (assume local source build)
    """
    # CI environments (building wheels for distribution)
    if os.environ.get("CI") in ("true", "1"):
        return True

    # cibuildwheel
    if os.environ.get("CIBUILDWHEEL"):
        return True

    # manylinux/auditwheel builds
    if os.environ.get("AUDITWHEEL_PLAT"):
        return True

    # Explicit marker from our workflows
    if os.environ.get("AUTOBAHN_WHEEL_BUILD") in ("true", "1"):
        return True

    # Default: assume local source build
    return False


def get_compile_args():
    """
    Returns appropriate compiler arguments for building NVX native extensions.

    This function determines the optimal compiler flags based on:
    - Target platform (Windows, Linux, macOS, etc.)
    - Target architecture (x86-64, ARM64, etc.)
    - Build context (wheel distribution vs. local source install)
    - User/distro overrides via environment variables

    Returns
    -------
    list of str
        Compiler arguments suitable for current build context.
        For MSVC: ['/O2', '/W3']
        For GCC/Clang: ['-std=c99', '-Wall', '-O3', '-march=...']

    Examples
    --------
    >>> args = get_compile_args()
    >>> # Use with CFFI:
    >>> ffi.set_source("_nvx_module", c_source, extra_compile_args=args)
    """
    if sys.platform == "win32":
        # MSVC on Windows
        # Note: MSVC doesn't support -march, uses /arch instead
        # But default codegen is already optimized for current arch
        return ["/O2", "/W3"]

    # GCC/Clang on POSIX (Linux, macOS, *BSD)
    machine = platform.machine().lower()

    # Base flags for all POSIX platforms
    base_args = [
        "-std=c99",
        "-Wall",
        "-Wno-strict-prototypes",
        "-O3",
    ]

    # User can explicitly override architecture target
    arch_override = os.environ.get("AUTOBAHN_ARCH_TARGET", "").lower()

    if arch_override == "native":
        # User explicitly wants -march=native (maximum performance)
        # Use case: Gentoo, Arch Linux, performance-critical deployments
        return base_args + ["-march=native"]

    elif arch_override == "safe":
        # User explicitly wants portable baseline
        # Use case: Debian, Ubuntu, RHEL package builds
        arch_flag = _get_safe_march_flag(machine)
        if arch_flag:
            return base_args + [arch_flag]
        else:
            # Unknown arch: use compiler defaults
            return base_args

    elif is_building_wheel():
        # Building wheel for distribution: use portable baseline
        # These wheels may run on CPUs different from build machine
        arch_flag = _get_safe_march_flag(machine)
        if arch_flag:
            return base_args + [arch_flag]
        else:
            # Unknown arch: use compiler defaults
            return base_args

    else:
        # Building from source locally: use -march=native for maximum performance
        # Build machine = runtime machine, so native optimizations are safe and optimal
        return base_args + ["-march=native"]


def _get_safe_march_flag(machine):
    """
    Returns safe -march flag for portable binaries on the given machine architecture.

    Parameters
    ----------
    machine : str
        Machine architecture from platform.machine().lower()

    Returns
    -------
    str or None
        Safe -march flag for the architecture, or None for unknown architectures.
    """
    if machine in ("x86_64", "amd64", "x64"):
        # x86-64: Use microarchitecture level 2 (2009+ CPUs)
        # Includes: SSE4.2, POPCNT, SSSE3, SSE4.1
        # Compatible with: Intel Nehalem+, AMD Bulldozer+
        return "-march=x86-64-v2"

    elif machine in ("aarch64", "arm64"):
        # ARM64: Use ARMv8-A baseline (all 64-bit ARM)
        # Compatible with: Raspberry Pi 4/5, AWS Graviton, Apple Silicon
        return "-march=armv8-a"

    else:
        # Unknown architecture: let compiler use safe defaults
        # (no -march flag = generic code for the architecture)
        return None



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Publish(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Publish()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPublish(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Publish
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Publish
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Publish
    def Request(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Publish
    def Topic(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Positional values for application-defined event payload.
    # Publish
    def Args(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Publish
    def ArgsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Publish
    def ArgsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def ArgsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Keyword values for application-defined event payload.
    # Publish
    def Kwargs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Publish
    def KwargsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Publish
    def KwargsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def KwargsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Alternative, transparent payload. If given, ``args`` and ``kwargs`` must be left unset.
    # Publish
    def Payload(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Publish
    def PayloadAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Publish
    def PayloadLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def PayloadIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # The specific scheme in use with Payload Passthru (PPT) mode for the application payload.
    # Publish
    def PptScheme(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The specific serializer encoding the application payload with the Payload Passthru (PPT) scheme in use.
    # Publish
    def PptSerializer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The cryptographic algorithm ("cipher") encrypting the application payload with the Payload Passthru (PPT) scheme in use.
    # Publish
    def PptCipher(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The identifier or reference to the encryption key that was used to encrypt the payload with the Payload Passthru (PPT) scheme and cipher in use.
    # Publish
    def PptKeyid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Publish
    def Acknowledge(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # Publish
    def ExcludeMe(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return True

    # Publish
    def Exclude(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8),
            )
        return 0

    # Publish
    def ExcludeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # Publish
    def ExcludeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def ExcludeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        return o == 0

    # Publish
    def ExcludeAuthid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4)
            )
        return ""

    # Publish
    def ExcludeAuthidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def ExcludeAuthidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        return o == 0

    # Publish
    def ExcludeAuthrole(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4)
            )
        return ""

    # Publish
    def ExcludeAuthroleLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def ExcludeAuthroleIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        return o == 0

    # Publish
    def Eligible(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8),
            )
        return 0

    # Publish
    def EligibleAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # Publish
    def EligibleLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def EligibleIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        return o == 0

    # Publish
    def EligibleAuthid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4)
            )
        return ""

    # Publish
    def EligibleAuthidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def EligibleAuthidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        return o == 0

    # Publish
    def EligibleAuthrole(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4)
            )
        return ""

    # Publish
    def EligibleAuthroleLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def EligibleAuthroleIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        return o == 0

    # Publish
    def Retain(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # Publish
    def TransactionHash(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Publish
    def ForwardFor(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from autobahn.wamp.gen.wamp.proto.Principal import Principal

            obj = Principal()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Publish
    def ForwardForLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publish
    def ForwardForIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        return o == 0


def PublishStart(builder):
    builder.StartObject(21)


def Start(builder):
    PublishStart(builder)


def PublishAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    PublishAddSession(builder, session)


def PublishAddRequest(builder, request):
    builder.PrependUint64Slot(1, request, 0)


def AddRequest(builder, request):
    PublishAddRequest(builder, request)


def PublishAddTopic(builder, topic):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(topic), 0
    )


def AddTopic(builder, topic):
    PublishAddTopic(builder, topic)


def PublishAddArgs(builder, args):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(args), 0
    )


def AddArgs(builder, args):
    PublishAddArgs(builder, args)


def PublishStartArgsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartArgsVector(builder, numElems):
    return PublishStartArgsVector(builder, numElems)


def PublishAddKwargs(builder, kwargs):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(kwargs), 0
    )


def AddKwargs(builder, kwargs):
    PublishAddKwargs(builder, kwargs)


def PublishStartKwargsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartKwargsVector(builder, numElems):
    return PublishStartKwargsVector(builder, numElems)


def PublishAddPayload(builder, payload):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(payload), 0
    )


def AddPayload(builder, payload):
    PublishAddPayload(builder, payload)


def PublishStartPayloadVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPayloadVector(builder, numElems):
    return PublishStartPayloadVector(builder, numElems)


def PublishAddPptScheme(builder, pptScheme):
    builder.PrependUint8Slot(6, pptScheme, 0)


def AddPptScheme(builder, pptScheme):
    PublishAddPptScheme(builder, pptScheme)


def PublishAddPptSerializer(builder, pptSerializer):
    builder.PrependUint8Slot(7, pptSerializer, 0)


def AddPptSerializer(builder, pptSerializer):
    PublishAddPptSerializer(builder, pptSerializer)


def PublishAddPptCipher(builder, pptCipher):
    builder.PrependUint8Slot(8, pptCipher, 0)


def AddPptCipher(builder, pptCipher):
    PublishAddPptCipher(builder, pptCipher)


def PublishAddPptKeyid(builder, pptKeyid):
    builder.PrependUOffsetTRelativeSlot(
        9, flatbuffers.number_types.UOffsetTFlags.py_type(pptKeyid), 0
    )


def AddPptKeyid(builder, pptKeyid):
    PublishAddPptKeyid(builder, pptKeyid)


def PublishAddAcknowledge(builder, acknowledge):
    builder.PrependBoolSlot(10, acknowledge, 0)


def AddAcknowledge(builder, acknowledge):
    PublishAddAcknowledge(builder, acknowledge)


def PublishAddExcludeMe(builder, excludeMe):
    builder.PrependBoolSlot(11, excludeMe, 1)


def AddExcludeMe(builder, excludeMe):
    PublishAddExcludeMe(builder, excludeMe)


def PublishAddExclude(builder, exclude):
    builder.PrependUOffsetTRelativeSlot(
        12, flatbuffers.number_types.UOffsetTFlags.py_type(exclude), 0
    )


def AddExclude(builder, exclude):
    PublishAddExclude(builder, exclude)


def PublishStartExcludeVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def StartExcludeVector(builder, numElems):
    return PublishStartExcludeVector(builder, numElems)


def PublishAddExcludeAuthid(builder, excludeAuthid):
    builder.PrependUOffsetTRelativeSlot(
        13, flatbuffers.number_types.UOffsetTFlags.py_type(excludeAuthid), 0
    )


def AddExcludeAuthid(builder, excludeAuthid):
    PublishAddExcludeAuthid(builder, excludeAuthid)


def PublishStartExcludeAuthidVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartExcludeAuthidVector(builder, numElems):
    return PublishStartExcludeAuthidVector(builder, numElems)


def PublishAddExcludeAuthrole(builder, excludeAuthrole):
    builder.PrependUOffsetTRelativeSlot(
        14, flatbuffers.number_types.UOffsetTFlags.py_type(excludeAuthrole), 0
    )


def AddExcludeAuthrole(builder, excludeAuthrole):
    PublishAddExcludeAuthrole(builder, excludeAuthrole)


def PublishStartExcludeAuthroleVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartExcludeAuthroleVector(builder, numElems):
    return PublishStartExcludeAuthroleVector(builder, numElems)


def PublishAddEligible(builder, eligible):
    builder.PrependUOffsetTRelativeSlot(
        15, flatbuffers.number_types.UOffsetTFlags.py_type(eligible), 0
    )


def AddEligible(builder, eligible):
    PublishAddEligible(builder, eligible)


def PublishStartEligibleVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def StartEligibleVector(builder, numElems):
    return PublishStartEligibleVector(builder, numElems)


def PublishAddEligibleAuthid(builder, eligibleAuthid):
    builder.PrependUOffsetTRelativeSlot(
        16, flatbuffers.number_types.UOffsetTFlags.py_type(eligibleAuthid), 0
    )


def AddEligibleAuthid(builder, eligibleAuthid):
    PublishAddEligibleAuthid(builder, eligibleAuthid)


def PublishStartEligibleAuthidVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartEligibleAuthidVector(builder, numElems):
    return PublishStartEligibleAuthidVector(builder, numElems)


def PublishAddEligibleAuthrole(builder, eligibleAuthrole):
    builder.PrependUOffsetTRelativeSlot(
        17, flatbuffers.number_types.UOffsetTFlags.py_type(eligibleAuthrole), 0
    )


def AddEligibleAuthrole(builder, eligibleAuthrole):
    PublishAddEligibleAuthrole(builder, eligibleAuthrole)


def PublishStartEligibleAuthroleVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartEligibleAuthroleVector(builder, numElems):
    return PublishStartEligibleAuthroleVector(builder, numElems)


def PublishAddRetain(builder, retain):
    builder.PrependBoolSlot(18, retain, 0)


def AddRetain(builder, retain):
    PublishAddRetain(builder, retain)


def PublishAddTransactionHash(builder, transactionHash):
    builder.PrependUOffsetTRelativeSlot(
        19, flatbuffers.number_types.UOffsetTFlags.py_type(transactionHash), 0
    )


def AddTransactionHash(builder, transactionHash):
    PublishAddTransactionHash(builder, transactionHash)


def PublishAddForwardFor(builder, forwardFor):
    builder.PrependUOffsetTRelativeSlot(
        20, flatbuffers.number_types.UOffsetTFlags.py_type(forwardFor), 0
    )


def AddForwardFor(builder, forwardFor):
    PublishAddForwardFor(builder, forwardFor)


def PublishStartForwardForVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartForwardForVector(builder, numElems):
    return PublishStartForwardForVector(builder, numElems)


def PublishEnd(builder):
    return builder.EndObject()


def End(builder):
    return PublishEnd(builder)
